@@ -4,9 +4,9 @@ from dataclasses import dataclass
 from typing import (Any, Callable, Generic, NoReturn, Optional, Type, TypeVar,
                     Union)
 
-from typing_extensions import Concatenate, ParamSpec, Protocol
-from ziopy.util import RaiseLeft, raise_exception
+from typing_extensions import Concatenate, Final, ParamSpec, Protocol
 
+from ziopy.util import RaiseLeft, raise_exception
 
 A = TypeVar('A', covariant=True)
 A1 = TypeVar('A1')
@@ -26,6 +26,7 @@ C2 = TypeVar('C2')
 E = TypeVar('E', covariant=True)
 E1 = TypeVar('E1')
 E2 = TypeVar('E2')
+E_con = TypeVar('E_con', contravariant=True)
 
 X = TypeVar('X', bound=BaseException)
 P = ParamSpec("P")
@@ -42,6 +43,11 @@ class TypeMatchException(Generic[AA], Exception):
     value: Optional[AA]
 
 
+class GetItem(Protocol[A_con, B]):
+    def __getitem__(self, item: A_con) -> B:
+        pass
+
+
 class Either(Generic[A, B], metaclass=ABCMeta):
     """Right-biased disjunction"""
 
@@ -66,6 +72,12 @@ class Either(Generic[A, B], metaclass=ABCMeta):
         else:
             raise TypeError()
 
+    @staticmethod
+    def from_optional(value: Optional[B]) -> "Either[None, B]":
+        if value is None:
+            return Either.left(value)
+        return Either.right(value)
+
     def to_left(self: "Either[AA, NoReturn]") -> "Left[AA]":
         if not isinstance(self, Left):
             raise TypeError("to_left can only be called on an instance of Left.")
@@ -179,6 +191,14 @@ class Either(Generic[A, B], metaclass=ABCMeta):
     def to_union(self) -> Union[A, B]:
         return self.match(lambda x: x.value, lambda y: y.value)
 
+    def cast(self: "Either[AA, BB]", t: Type[C]) -> "Either[Union[AA, TypeError], C]":
+        return self.flat_map(
+            lambda b: Either.right(b) if isinstance(b, t)
+            else Either.left(
+                TypeError(f"Unable to cast value {b} (of type {type(b).__name__}) as type {t.__name__}.")
+            )
+        )
+
 
 @dataclass(frozen=True)
 class Left(Generic[A], Either[A, NoReturn]):
@@ -190,8 +210,8 @@ class Right(Generic[B], Either[NoReturn, B]):
     value: B
 
 
-class EitherMonad(Generic[E1]):
-    def __lshift__(self, arg: Either[E1, BB]) -> BB:
+class EitherMonad(Generic[E_con]):
+    def __lshift__(self, arg: Either[E_con, BB]) -> BB:
         return arg.fold(
             lambda e: raise_exception(RaiseLeft(e)),
             lambda a: a
@@ -233,6 +253,11 @@ class WithArgs(Protocol[P, A_con, B]):
         pass
 
 
+class _Identities:
+    def __getitem__(self, t: Type[AA]) -> "EitherArrow[[AA], NoReturn, AA]":
+        return EitherArrow.from_callable(lambda x: x)
+
+
 class EitherArrow(Generic[P, A, B]):
     @staticmethod
     def from_either(either: Either[E, Callable[P, A]]) -> "EitherArrow[P, E, A]":
@@ -242,6 +267,16 @@ class EitherArrow(Generic[P, A, B]):
     def from_callable(f: Callable[P, A]) -> "EitherArrow[P, NoReturn, A]":
         return EitherArrow(lambda *args, **kwargs: Either.right(f(*args, **kwargs)))
 
+    @staticmethod
+    def right(x: AA) -> "EitherArrow[[], NoReturn, AA]":
+        return EitherArrow(lambda: Either.right(x))
+
+    @staticmethod
+    def left(x: BB) -> "EitherArrow[[], BB, NoReturn]":
+        return EitherArrow(lambda: Either.left(x))
+
+    identity: Final = _Identities()
+
     def __init__(self, f: Callable[P, Either[A, B]]) -> None:
         self._f = f
 
@@ -340,6 +375,39 @@ class EitherArrow(Generic[P, A, B]):
     def to_union(self) -> Callable[P, Union[A, B]]:
         return lambda *args, **kwargs: self._f(*args, **kwargs).to_union()
 
+    def cast(self: "EitherArrow[P, AA, BB]", t: Type[C]) -> "EitherArrow[P, Union[AA, TypeError], C]":
+        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).cast(t))
+
+    def __getitem__(self: "EitherArrow[P, AA, GetItem[B1, C1]]", key: B1) -> "EitherArrow[P, AA, C1]":
+        return EitherArrow(
+            lambda *args, **kwargs: self._f(*args, **kwargs).map(lambda x: x[key])
+        )
+
+    def match_types(self: "EitherArrow[P, AA, BB]") -> "EitherArrow[P, AA, NoReturn]":
+        def _f(arg: BB) -> NoReturn:
+            raise TypeMatchException(arg)
+        return self.map(_f)
+
+    def at_type(
+        self: "EitherArrow[P, AA, BB]",
+        target_type: Type[B2],
+        operation: "Callable[[B2], Either[A2, C]]"
+    ) -> "EitherArrow[P, Union[AA, A2], Union[BB, C]]":
+        def _recover(arg: Union[BB, TypeMatchException]) -> Either[A2, Union[BB, C]]:
+            if not isinstance(arg, TypeMatchException):
+                return Either.right(arg)
+            if isinstance(arg.value, target_type):
+                return operation(arg.value)
+            else:
+                raise arg from arg
+
+        return (
+            self.swap()
+            .catch(TypeMatchException)
+            .swap()
+            .flat_map(_recover)
+        )
+
     def with_args_map(
         self: "EitherArrow[P, AA, BB]",
         f: WithArgs[P, BB, C]
